import copy
import logging
from typing import Any, TypeVar

from fastapi import Request
from jose import JWTError, jwt

from coderfastapi.lib.security.policies.authentication import AuthenticationPolicy
from coderfastapi.lib.security.policies.authentication.jwt.providers import (
    JWTDataProvider,
)
from coderfastapi.lib.security.policies.authentication.jwt.providers.expiration import (
    ExpirationDataProvider,
)
from coderfastapi.lib.security.policies.authentication.jwt.providers.recovery import (
    RecoveryDataProvider,
)
from coderfastapi.lib.security.policies.authentication.jwt.providers.user import (
    UserDataProvider,
)

log = logging.getLogger(__name__)
T = TypeVar("T", bound=Request)


class JWTAuthenticationPolicy(AuthenticationPolicy):
    secret_key: str
    algorithm: str
    providers: set[JWTDataProvider]

    def __init__(
        self,
        secret_key: str,
        algorithm: str = "HS256",
        providers: set[JWTDataProvider] = {
            ExpirationDataProvider(),
            RecoveryDataProvider(),
            UserDataProvider(),
        },
    ) -> None:
        self.secret_key = secret_key
        self.algorithm = algorithm
        self.providers = providers

    def authenticate_request(self, request: T) -> T:
        request_ = copy.copy(request)
        decoded_data = self._decode_access_token(request_.headers)
        for provider in self.providers:
            request_ = provider.augment_request(request_, decoded_data)
        return request_

    def _decode_access_token(self, headers: dict[str, Any]) -> dict[str, Any]:
        try:
            auth_method, access_token = get_auth_method_and_token(
                headers["authorization"]
            )
            if auth_method != "Bearer":
                return {}

            return jwt.decode(
                access_token,
                self.secret_key,
                algorithms=[self.algorithm],
            )
        except (JWTError, KeyError, ValueError):
            return {}

    def create_access_token(self, **kwargs) -> str:
        data = {}
        for provider in self.providers:
            data.update(provider.parse_to_encode(kwargs))
        return jwt.encode(
            data,
            self.secret_key,
            algorithm=self.algorithm,
        )


def get_auth_method_and_token(
    authorization_header: str,
) -> tuple[str, str] | tuple[None, None]:
    try:
        auth_method, token_string = authorization_header.split(" ")
        return auth_method, token_string
    except (AttributeError, ValueError):
        return None, None



import logging

from fastapi import Request

from coderfastapi.lib.security import Authenticated
from coderfastapi.lib.security.policies.authorization import AuthorizationPolicy

log = logging.getLogger(__name__)


class UserAuthorizationPolicy(AuthorizationPolicy):
    @classmethod
    def get_principals(cls, request: Request) -> tuple[str, ...]:
        principals = super().get_principals(request)

        authenticated_user_id = request.user_id
        if authenticated_user_id:
            principals = cls._with_authenticated_user_principals(
                principals,
                str(authenticated_user_id),
            )

        log.debug(f"Found principals: {principals}")

        return principals

    @staticmethod
    def _with_authenticated_user_principals(
        principals: tuple[str, ...],
        authenticated_user_id: str,
    ) -> tuple[str, ...]:
        return principals + (Authenticated, f"user:{authenticated_user_id}")


class RecoverableUserAuthorizationPolicy(UserAuthorizationPolicy):
    @classmethod
    def get_principals(cls, request: Request) -> tuple[str, ...]:
        if not request.recovery:
            return super().get_principals(request)

        principals = AuthorizationPolicy.get_principals(request) + (
            f"recovering_user:{request.user_id}",
        )
        log.debug(f"Found principals: {principals}")
        return principals



import logging
from typing import Any

from coderfastapi.lib.security.policies.authentication.jwt.providers import (
    JWTDataProvider,
    T,
)

log = logging.getLogger(__name__)


class RecoveryDataProvider(JWTDataProvider):
    def augment_request(self, request: T, data: dict[str, Any]) -> T:
        request_ = super().augment_request(request, data)
        request_.recovery = data.get("recovery", False)
        log.info(f"Recovery mode: {request_.recovery}")
        return request_

    def parse_to_encode(self, data: dict[str, Any]) -> dict[str, Any]:
        return {"recovery": data.get("recovery", False)}



from abc import ABC, abstractmethod
from typing import TypeVar

from fastapi import Request

T = TypeVar("T", bound=Request)


class AuthenticationPolicy(ABC):
    @abstractmethod
    def authenticate_request(self, request: T) -> T:
        """Adds metadata to request if properly authenticated."""



import copy
import inspect
from typing import Any, Awaitable, Callable, ParamSpec, TypeVar

from fastapi import APIRouter, Request

from coderfastapi.lib.security.acl import ACLProvider
from coderfastapi.lib.security.policies.authentication import AuthenticationPolicy
from coderfastapi.lib.security.policies.authorization import AuthorizationPolicy
from coderfastapi.lib.signature import copy_parameters

T = TypeVar("T")
P = ParamSpec("P")


class SecureRouter(APIRouter):
    authentication_policy: AuthenticationPolicy
    authorization_policy: AuthorizationPolicy

    def __init__(
        self,
        authentication_policy: AuthenticationPolicy,
        authorization_policy: AuthorizationPolicy,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.authentication_policy = authentication_policy
        self.authorization_policy = authorization_policy

    async def _call_handler_with_authentication(
        self,
        handler: Callable[P, Awaitable[T] | T],
        permission: str,
        request: Request,
        context: Any | None,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T:
        context_acl_provider = None
        if isinstance(context, ACLProvider):
            context_acl_provider = context

        authenticated_request = self.authentication_policy.authenticate_request(request)
        self.authorization_policy.validate_permission(
            permission,
            authenticated_request,
            context_acl_provider,
        )

        kwargs = self._propagate_params(handler, kwargs, authenticated_request, context)
        output = handler(*args, **kwargs)
        if inspect.iscoroutine(output):
            output = await output

        return output

    @staticmethod
    def _propagate_params(
        func: Callable[P, Awaitable[T] | T],
        kwargs: [str, Any],
        request: Request,
        context: Any | None,
    ) -> dict[str, Any]:
        new_kwargs = copy.copy(kwargs)
        func_signature = inspect.signature(func)
        if func_signature.parameters.get("context"):
            new_kwargs["context"] = context
        if func_signature.parameters.get("request"):
            new_kwargs["request"] = request
        return new_kwargs

    def post(self, *outer_args, **outer_kwargs):
        return lambda func: self._http_method(func, "post", *outer_args, **outer_kwargs)

    def put(self, *outer_args, **outer_kwargs):
        return lambda func: self._http_method(func, "put", *outer_args, **outer_kwargs)

    def patch(self, *outer_args, **outer_kwargs):
        return lambda func: self._http_method(
            func, "patch", *outer_args, **outer_kwargs
        )

    def delete(self, *outer_args, **outer_kwargs):
        return lambda func: self._http_method(
            func, "delete", *outer_args, **outer_kwargs
        )

    def get(self, *outer_args, **outer_kwargs):
        return lambda func: self._http_method(func, "get", *outer_args, **outer_kwargs)

    def options(self, *outer_args, **outer_kwargs):
        return lambda func: self._http_method(
            func, "options", *outer_args, **outer_kwargs
        )

    def trace(self, *outer_args, **outer_kwargs):
        return lambda func: self._http_method(
            func, "trace", *outer_args, **outer_kwargs
        )

    def head(self, *outer_args, **outer_kwargs):
        return lambda func: self._http_method(func, "head", *outer_args, **outer_kwargs)

    def _http_method(
        self,
        func: Callable[P, Awaitable[T] | T],
        http_method: str,
        *outer_args,
        permission: str = "public",
        **outer_kwargs,
    ) -> Callable[[Request, P], Awaitable[T]]:
        route = getattr(super(SecureRouter, self), http_method)

        @route(*outer_args, **outer_kwargs)
        async def wrapped(
            request: Request,
            context: Any | None = None,
            *args,
            **kwargs,
        ) -> T:
            return await self._call_handler_with_authentication(
                func, permission, request, context, *args, **kwargs
            )

        wrapped_signature = inspect.signature(wrapped)
        func_signature = inspect.signature(func)
        wrapped.__signature__ = copy_parameters(
            wrapped_signature, func_signature, ["request"]
        )
        return wrapped



from datetime import datetime
from typing import Any

from coderfastapi.lib.security.policies.authentication.jwt.providers import (
    JWTDataProvider,
)


class ExpirationDataProvider(JWTDataProvider):
    def parse_to_encode(self, data: dict[str, Any]) -> dict[str, Any]:
        return {"exp": datetime.utcnow() + data["delta"]}



from functools import wraps
from http import HTTPStatus
from typing import Awaitable, Callable, Optional, TypeVar

from fastapi import HTTPException

T = TypeVar("T")


def http_require(
    entity_name: str,
) -> Callable[[Callable[..., Awaitable[Optional[T]]]], Callable[..., Awaitable[T]]]:
    def decorate(
        func: Callable[..., Awaitable[Optional[T]]]
    ) -> Callable[..., Awaitable[T]]:
        @wraps(func)
        async def wrapper(*args, **kwargs) -> T:
            entity = await func(*args, **kwargs)
            if entity is None:
                raise HTTPException(
                    status_code=HTTPStatus.NOT_FOUND,
                    detail=f"The {entity_name} is not found.",
                )
            return entity

        return wrapper

    return decorate



import copy
from abc import ABC, abstractmethod
from typing import Any, TypeVar

from fastapi import Request

T = TypeVar("T", bound=Request)


class JWTDataProvider(ABC):
    def augment_request(self, request: T, data: dict[str, Any]) -> T:
        return copy.copy(request)

    @abstractmethod
    def parse_to_encode(self, data: dict[str, Any]) -> dict[str, Any]:
        """Returns data that needs to be encoded in the JWT for this provider."""
